# This file is placed in the Public Domain.


"big Object."


import copy as copying
import datetime
import json
import os
import pathlib
import sys
import time
import types
import uuid


## defines


def __dir__():
    return (
        "Object",
        "clear",
        "copy",
        "fromkeys",
        "get",
        "items",
        "keys",
        "pop",
        "popitem",
        "setdefault",
        "update",
        "values",
        "ObjectDecoder",
        "ObjectEncoder",
        "dump",
        "dumps",
        "load",
        "loadp",
        "loads",
        "save",
        "savep",
        "diff",
        "edit",
        "fmt",
        "getname",
        "gettype",
        "idx",
        "keys",
        "register",
        "search",
        "set"
    )


## exceptions


class NoModule(Exception):

     pass


class NoType(Exception):

     pass


## big Object


class Object:

    __slots__ = (
        "__dict__",
        "__idx__",
        "__otype__",
        "__stp__",
    )

    def __init__(self):
        super().__init__()
        self.__otype__ = str(type(self)).split()[-1][1:-2]
        self.__stp__ = os.path.join(
            self.__otype__,
            str(uuid.uuid4()),
            os.sep.join(str(datetime.datetime.now()).split()),
        )
        self.__idx__ = 0
        
    def __contains__(self, k):
        if k in keys(self):
            return True
        return False

    def __delitem__(self, k):
        if k in self:
            del self.__dict__[k]

    def __eq__(self, o):
        return len(self) == len(o)

    def __getitem__(self, k):
        return self.__dict__[k]

    def __iter__(self):
        return iter(self.__dict__)

    def __len__(self):
        return len(self.__dict__)

    def __le__(self, o):
        return len(self) <= len(o)

    def __lt__(self, o):
        return len(self) < len(o)

    def __ge__(self, o):
        return len(self) >= len(o)

    def __gt__(self, o):
        return len(self) > len(o)

    def __hash__(self):
        return id(self)

    def __reduce__(self):
        pass

    def __reduce_ex__(self, k):
        pass

    def __setitem__(self, k, v):
        self.__dict__[k] = v

    def __oqn__(self):
        return "<%s.%s object at %s>" % (
            self.__class__.__module__,
            self.__class__.__name__,
            hex(id(self)),
        )

    def __str__(self):
        return str(self.__dict__)


## object interface


def clear(o):
    o.__dict__ = {}


def copy(o):
    return copying.copy(o)


def fromkeys(iterable, value=None):
    o = Object()
    for i in iterable:
        o[i] = value    
    return o


def get(o, key, default=None):
    return o.__dict__.get(key, default)


def items(o):
    try:
        return o.__dict__.items()
    except AttributeError:
        return o.items()


def keys(o):
    try:
        return o.__dict__.keys()
    except TypeError:
        return o.keys()


def pop(o, k, d=None):
    try:
        return o[k]
    except KeyError:
        if d:
            return d
        raise KeyError


def popitem(o):
    k = keys(o)
    if k:
        v = o[k]
        del o[k]
        return (k, v)
    raise KeyError


def setdefault(o, key, default=None):
    if key not in o:
        o[key] = default
        return default
    else:
        return o[key]         

 
def update(o, data):
    try:
        o.__dict__.update(vars(data))
    except TypeError:
        o.__dict__.update(data)
    return o


def values(o):
    try:
        return o.__dict__.values()
    except TypeError:
        return o.values()


## object function


def diff(o1, o2):
    d = Object()
    for k in keys(o2):
        if k in keys(o1) and o1[k] != o2[k]:
            d[k] = o2[k]
    return d


def edit(o, setter, skip=True, skiplist=None):
    if skiplist is None:
        skiplist = []
    count = 0
    for key, v in items(setter):
        if skip and v == "":
            del o[key]
        if key in skiplist:
            continue
        count += 1
        if v in ["True", "true"]:
            o[key] = True
        elif v in ["False", "false"]:
            o[key] = False
        else:
            o[key] = v
    return count


def fmt(o, keyz=None, empty=True, skip=None, newline=False):
    if keyz is None:
        keyz = keys(o) or ["txt"]
    if skip is None:
        skip = []
    res = []
    txt = ""
    for key in keyz:
        if key in skip:
            continue
        if key in dir(o):
            if key.startswith("__"):
                continue
            val = getattr(o, key, None)
            if empty and not val:
                continue
            val = str(val).strip()
            res.append((key, val))
    result = []
    for k, v in res:
        if newline:
            result.append("%s=%s%s" % (k, v, "\n"))
        else:
            result.append("%s=%s%s" % (k, v, " "))
    if newline:
        txt += "\n".join([x.strip() for x in result])
    else:
        txt += " ".join([x.strip() for x in result])
    return txt


def getname(o):
    t = type(o)
    if isinstance(t, types.ModuleType):
        return o.__name__
    if "__self__" in dir(o):
        return "%s.%s" % (o.__self__.__class__.__name__, o.__name__)
    if "__class__" in dir(o) and "__name__" in dir(o):
        return "%s.%s" % (o.__class__.__name__, o.__name__)
    if "__class__" in dir(o):
        return o.__class__.__name__
    if "__name__" in dir(o):
        return o.__name__
    return None


def gettype(o):
    return str(type(o)).split()[-1][1:-2]


def idx(o, txt):
    o[o.__idx__] = txt
    o.__idx__ += 1


def register(o, k, v):
    o.__dict__[str(k)] = v


def search(o, s):
    ok = False
    for k, v in items(s):
        vv = getattr(o, k, None)
        if v not in str(vv):
            ok = False
            break
        ok = True
    return ok


def set(o, key, value):
    o.__dict__[key] = value


## configuration


class Cfg(Object):

    wd = ".obj"


## json


class ObjectDecoder(json.JSONDecoder):

    @staticmethod
    def decode(s, _w=None):
        o = Object()
        v = json.loads(s)
        update(o, v)
        return o


class ObjectEncoder(json.JSONEncoder):

    def default(self, o):
        if isinstance(o, dict):
            return o.items()
        if isinstance(o, Object):
            return vars(o)
        if isinstance(o, list):
            return iter(o)
        if isinstance(o,
                      (type(str), type(True), type(False),
                       type(int), type(float))):
            return o
        try:
            return json.JSONEncoder.default(self, o)
        except TypeError:
            return repr(o)


def dump(o, f):
    return json.dump(o, f, cls=ObjectEncoder)


def dumps(o):
    return json.dumps(o, cls=ObjectEncoder)


def load(o, opath):
    if opath.count(os.sep) != 3:
        return
    assert Cfg.wd
    splitted = opath.split(os.sep)
    stp = os.sep.join(splitted[-4:])
    lpath = os.path.join(Cfg.wd, "store", stp)
    if os.path.exists(lpath):
        with open(lpath, "r") as ofile:
            d = json.load(ofile, cls=ObjectDecoder)
            update(o, d)
    o.__stp__ = stp


def loads(s):
    return json.loads(s, cls=ObjectDecoder)


def loadp(o, opath):
    assert Cfg.wd
    if os.path.exists(opath):
        with open(opath, "r") as ofile:
            d = json.load(ofile, cls=ObjectDecoder)
            update(o, d)


def save(o):
    assert Cfg.wd
    prv = os.sep.join(o.__stp__.split(os.sep)[:2])
    o.__stp__ = os.path.join(prv,
                             os.sep.join(str(datetime.datetime.now()).split()))
    opath = os.path.join(Cfg.wd, "store", o.__stp__)
    savep(o, opath)
    with open(opath, "w") as ofile:
        json.dump(
            o.__dict__, ofile, cls=ObjectEncoder, indent=4, sort_keys=True
        )
    os.chmod(opath, 0o444)
    return o.__stp__


def savep(o, opath):
    cdir(opath)
    with open(opath, "w") as ofile:
        json.dump(
            o.__dict__, ofile, cls=ObjectEncoder, indent=4, sort_keys=True
        )
    return o.__stp__


def cdir(path):
    if os.path.exists(path):
        return
    if path.split(os.sep)[-1].count(":") == 2:
        path = os.path.dirname(path)
    pathlib.Path(path).mkdir(parents=True, exist_ok=True)


## database


class Db(Object):

    def all(self, otype, selector=None, index=None, timed=None):
        nr = -1
        if selector is None:
            selector = {}
        for fn in fns(otype, timed):
            o = hook(fn)
            if selector and not search(o, selector):
                continue
            if "_deleted" in o and o._deleted:
                continue
            nr += 1
            if index is not None and nr != index:
                continue
            yield fn, o

    def deleted(self, otype):
        for fn in fns(otype):
            o = hook(fn)
            if "_deleted" not in o or not o._deleted:
                continue
            yield fn, o

    def every(self, selector=None, index=None, timed=None):
        if selector is None:
            selector = {}
        nr = -1
        for otype in os.listdir(os.path.join(Cfg.wd, "store")):
            for fn in fns(otype, timed):
                o = hook(fn)
                if selector and not search(o, selector):
                    continue
                if "_deleted" in o and o._deleted:
                    continue
                nr += 1
                if index is not None and nr != index:
                    continue
                yield fn, o

    def find(self, otype, selector=None, index=None, timed=None):
        if selector is None:
            selector = {}
        got = False
        nr = -1
        for fn in fns(otype, timed):
            o = hook(fn)
            if selector and not search(o, selector):
                continue
            if "_deleted" in o and o._deleted:
                continue
            nr += 1
            if index is not None and nr != index:
                continue
            got = True
            yield (fn, o)
        if got:
            return (None, None)
        return None

    def lastmatch(self, otype, selector=None, index=None, timed=None):
        db = Db()
        res = sorted(db.find(otype, selector, index, timed),
                     key=lambda x: fntime(x[0]))
        if res:
            return res[-1]
        return (None, None)

    def lastobject(self, o):
        return self.lasttype(o.__otype__)

    def lasttype(self, otype):
        fnn = fns(otype)
        if fnn:
            return hook(fnn[-1])
        return None

    def lastfn(self, otype):
        fn = fns(otype)
        if fn:
            fnn = fn[-1]
            return (fnn, hook(fnn))
        return (None, None)


def all(timed=None):
    assert Cfg.wd
    p = os.path.join(Cfg.wd, "store")
    for name in os.listdir(p):
        for fn in fns(name):
            yield fn


def find(name, selector=None, index=None, timed=None, names=None):
    db = Db()
    for n in get(names or Object(), name, [name,]):
        for fn, o in db.find(n, selector, index, timed):
            yield fn, o

def last(o):
    db = Db()
    path, obj = db.lastfn(o.__otype__)
    if obj:
        update(o, obj)
    if path:
        splitted = path.split(os.sep)
        stp = os.sep.join(splitted[-4:])
        return stp
    return None


def fns(name, timed=None):
    if not name:
        return []
    assert Cfg.wd
    p = os.path.join(Cfg.wd, "store", name) + os.sep
    res = []
    d = ""
    for rootdir, dirs, _files in os.walk(p, topdown=False):
        if dirs:
            d = sorted(dirs)[-1]
            if d.count("-") == 2:
                dd = os.path.join(rootdir, d)
                fls = sorted(os.listdir(dd))
                if fls:
                    p = os.path.join(dd, fls[-1])
                    if (
                        timed
                        and "from" in timed
                        and timed["from"]
                        and fntime(p) < timed["from"]
                    ):
                        continue
                    if timed and timed.to and fntime(p) > timed.to:
                        continue
                    res.append(p)
    return sorted(res, key=fntime)


def fntime(daystr):
    daystr = daystr.replace("_", ":")
    datestr = " ".join(daystr.split(os.sep)[-2:])
    if "." in datestr:
        datestr, rest = datestr.rsplit(".", 1)
    else:
        rest = ""
    t = time.mktime(time.strptime(datestr, "%Y-%m-%d %H:%M:%S"))
    if rest:
        t += float("." + rest)
    else:
        t = 0
    return t


def hook(hfn, tbl=None):
    if hfn.count(os.sep) > 3:
        oname = hfn.split(os.sep)[-4:]
    else:
        oname = hfn.split(os.sep)
    cname = oname[0]
    fn = os.sep.join(oname)
    mod = None
    mn, cn = cname.rsplit(".", 1)
    if tbl:
        mod = tbl.get(mn)
    if not mod:
        mod = sys.modules.get(mn, None)
    if not mod:
        raise NoModule(mn)
    t = getattr(mod, cn, None)
    if t:
        o = t()
        load(o, fn)
        return o
    raise NoType(cname)


def listfiles(workdir):
    path = os.path.join(Cfg.wd, "store")
    if not os.path.exists(path):
        return []
    return sorted(os.listdir(path))
